//  
// Proteus 2012 lawnmower-man@mail.ru

#include<stdlib.h>
#include<assert.h>

class RandTree{
	struct Node{
		int v;
		Node *p1,*p2;
		int count;
		Node(int v):v(v),count(1),p1(0),p2(0) {}
		int Count() const {return this?count:0;}
	};
	Node *Fix(Node *node) {
		node->count=1+node->p1->Count()+node->p2->Count();
		return node;
	}
	Node *root;
	Node *Insert(Node*,int);
	Node *InsertRoot(Node*,int);
	Node *Remove(Node*,int);
	Node *Join(Node*,Node*);
	void Clear(Node*);
#ifdef _DEBUG
	int Check(Node *node)
	{
		if(!node) return 0;
		int count=1+Check(node->p1)+Check(node->p2);
		assert(count==node->count);
		return count;
	}
#endif
public:
	RandTree():root(0) {}
	~RandTree() {Clear(root);}
	bool Find(int);
	int  Size() {return root->Count();}
	void Insert(int v) {root=Insert(root,v);}
	void Remove(int v) {root=Remove(root,v);}
};

void RandTree::Clear(Node *node)
{
	if(node) Clear(node->p1),Clear(node->p2);
	delete node;
}

bool RandTree::Find(int v)
{
	for(Node *node=root; node;) {
		if(node->v==v) return true;
		node=(v<node->v)?node->p1:node->p2;
	}
	return false;
}


RandTree::Node *RandTree::Insert(Node *node,int v)
{
	if(!node || rand()%(node->count+1)==0) return InsertRoot(node,v); 
	if(v<node->v) node->p1=Insert(node->p1,v); 
	else node->p2=Insert(node->p2,v); 
	node->count++;
	return node;
}


RandTree::Node *RandTree::InsertRoot(Node *node, int v)
{
	Node *p;
	if(!node) return new Node(v); 
	if(v<node->v) {
		p=InsertRoot(node->p1,v); 
		node->p1=p->p2,p->p2=node; 
	} else {
		p=InsertRoot(node->p2,v);
		node->p2=p->p1,p->p1=node;
	}
	p->count=node->count+1;
	Fix(node);
	return p;
}


RandTree::Node *RandTree::Join(Node *p1,Node *p2)
{
	if(!p1) return p2;
	if(!p2) return p1;
	//if(!p1 || !p2) return (Node*)((long)p1|(long)p2);
	if(rand()%(p1->count+p2->count) < p1->count) {
		p1->p2=Join(p1->p2,p2); 
		return Fix(p1); 
	} else {
		p2->p1=Join(p1,p2->p1); 
		return Fix(p2);
	}
	return 0;
}


RandTree::Node *RandTree::Remove(Node *node,int v)
{
	if(!node) return 0; 
	if(v<node->v) node->p1=Remove(node->p1,v); 
	else if(v>node->v) node->p2=Remove(node->p2,v); 
	else {
		Node *d=node;
		node=Join(node->p1,node->p2); 
		delete d;
		if(!node) return node;
	}
	return Fix(node);
}



void main()
{
	int n;
	RandTree tree;
	for(n=0; n<100000; n++) tree.Insert(rand());
	for(n=0; n<100000; n++) tree.Remove(rand());
}







